# -*- coding: utf-8 -*-

import logging

log = logging.getLogger(__name__)

from queue import Queue
import signal
from threading import Event

from music2storage.connection import ConnectionHandler
from music2storage.helpers import convert_to_mp3, delete_local_file
from music2storage.signalhandler import SignalHandler
from music2storage.worker import Worker


class Music2Storage:
    """Manages workers, queues, services for music2storage."""

    def __init__(self):
        """Initializes all the queues and sets default values for services and workers."""

        self.queues = {
            'download': Queue(),
            'convert': Queue(),
            'upload': Queue(),
            'delete': Queue(),
            'done': Queue(),
        }

        self.connection_handler = ConnectionHandler()
        self.workers = []
        self.stopper = Event()
        self.signal_handler = None

    def add_to_queue(self, url):
        """
        Adds an URL to the download queue.

        :param str url: URL to the music service track
        """

        if self.connection_handler.current_music is None:
            log.error('Music service is not initialized. URL was not added to queue.')
        elif self.connection_handler.current_storage is None:
            log.error('Drive service is not initialized. URL was not added to queue.')
        else:
            self.queues['download'].put(url)

    def use_music_service(self, service_name, api_key=None):
        """
        Sets the current music service to service_name.
        
        :param str service_name: Name of the music service
        :param str api_key: Optional API key if necessary
        """

        self.connection_handler.use_music_service(service_name, api_key=api_key)

    def use_storage_service(self, service_name, custom_path=None):
        """
        Sets the current storage service to service_name and attempts to connect to it.
        
        :param str service_name: Name of the storage service
        :param str custom_path: Custom path where to download tracks for local storage (optional, and must already exist, use absolute paths only)
        """

        self.connection_handler.use_storage_service(service_name, custom_path=custom_path)

    def start_workers(self, workers_per_task=1):
        """
        Creates and starts the workers, as well as attaching a handler to terminate them gracefully when a SIGINT signal is received.

        :param int workers_per_task: Number of workers to create for each task in the pipeline
        """

        if not self.workers:
            for _ in range(workers_per_task):
                self.workers.append(Worker(self._download, self.queues['download'], self.queues['convert'], self.stopper))
                self.workers.append(Worker(self._convert, self.queues['convert'], self.queues['upload'], self.stopper))
                self.workers.append(Worker(self._upload, self.queues['upload'], self.queues['delete'], self.stopper))
                self.workers.append(Worker(self._delete, self.queues['delete'], self.queues['done'], self.stopper))

            self.signal_handler = SignalHandler(self.workers, self.stopper)
            signal.signal(signal.SIGINT, self.signal_handler)

            for worker in self.workers:
                worker.start()

    def _download(self, url):
        """
        Downloads the file associated with the given URL.

        :param str url: URL of the video/audio file
        :return str: Filename of the file in local storage
        """
        
        return self.connection_handler.current_music.download(url)

    def _convert(self, file_name):
        """
        Converts the file associated with the file_name passed into a MP3 file.

        :param str file_name: Filename and extension of the file to be converted
        :return str: Filename of the new file in local storage
        """

        return convert_to_mp3(file_name, self.queues['delete'])

    def _upload(self, file_name):
        """
        Uploads the file associated with the file_name passed to the storage service.

        :param str file_name: Filename of the file to be uploaded
        :return str: Original filename passed as an argument
        """

        return self.connection_handler.current_storage.upload(file_name)

    def _delete(self, file_name):
        """
        Deletes the file associated with the file_name passed from local storage.

        :param str file_name: Filename of the file to be deleted
        :return str: Filename of the file that was just deleted
        """

        return delete_local_file(file_name)

    def _notify(self):
        """TODO: Notify that the MP3 has been uploaded through a messaging service."""
        pass



# -*- coding: utf-8 -*-

from music2storage import log
from music2storage.service import Youtube, Soundcloud, GoogleDrive, LocalStorage


class ConnectionHandler:
    """Manages all the API connections to the music and storage services."""

    def __init__(self):
        self.current_music = None
        self.current_storage = None
        self.music_services = {}
        self.storage_services = {}

    def use_music_service(self, service_name, api_key):
        """
        Sets the current music service to service_name.

        :param str service_name: Name of the music service
        :param str api_key: Optional API key if necessary
        """

        try:
            self.current_music = self.music_services[service_name]
        except KeyError:
            if service_name == 'youtube':
                self.music_services['youtube'] = Youtube()
                self.current_music = self.music_services['youtube']
            elif service_name == 'soundcloud':
                self.music_services['soundcloud'] = Soundcloud(api_key=api_key)
                self.current_music = self.music_services['soundcloud']
            else:
                log.error('Music service name is not recognized.')

    def use_storage_service(self, service_name, custom_path):
        """
        Sets the current storage service to service_name and runs the connect method on the service.

        :param str service_name: Name of the storage service
        :param str custom_path: Custom path where to download tracks for local storage (optional, and must already exist, use absolute paths only)
        """

        try:
            self.current_storage = self.storage_services[service_name]
        except KeyError:
            if service_name == 'google drive':
                self.storage_services['google drive'] = GoogleDrive()
                self.current_storage = self.storage_services['google drive']
                self.current_storage.connect()
            elif service_name == 'dropbox':
                log.error('Dropbox is not supported yet.')
            elif service_name == 'local':
                self.storage_services['local'] = LocalStorage(custom_path=custom_path)
                self.current_storage = self.storage_services['local']
                self.current_storage.connect()
            else:
                log.error('Storage service name is not recognized.')



# -*- coding: utf-8 -*-

import sys

from music2storage import log


class SignalHandler:
    """Handles signals and stops the worker threads."""

    def __init__(self, workers, stopper):
        """
        Creates a SignalHandler associated with all the worker threads and a stopper Event object.

        :param list workers: List of all the worker threads from the Worker class
        :param threading.Event stopper: Event that signals that the thread should stop execution
        """

        self.workers = workers
        self.stopper = stopper

    def __call__(self, signum, frame):
        """
        Gets called by Python's signal module when the threads should be stopped.

        :param signal signum: Signal number that must be handled
        :param Frame frame: Current stack frame
        """

        log.info('Gracefully killing the threads...')
        self.stopper.set()

        for worker in self.workers:
            worker.join(1)

        sys.exit(0)



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
import os
import sys
from time import time

from apiclient.discovery import build
from googleapiclient.http import MediaFileUpload
from httplib2 import Http
from oauth2client import file, client, tools
from oauth2client.clientsecrets import InvalidClientSecretsError
from pytube import YouTube
from pytube.exceptions import RegexMatchError
import requests
from requests.exceptions import HTTPError
import soundcloud
from tqdm import tqdm

from music2storage import log


class MusicService(ABC):
    """Template for every music service."""

    @abstractmethod
    def download(self, url):
        """Downloads a song file from the music service."""


class StorageService(ABC):
    """Template for every storage service."""

    @abstractmethod
    def connect(self):
        """Connects to the storage service, sets connection attribute."""

    @abstractmethod
    def upload(self, file_name):
        """Uploads a file to the storage."""


class Youtube(MusicService):
    """Youtube service class."""

    def __init__(self):
        self.name = 'youtube'

    def download(self, url):
        """
        Downloads a MP4 or WebM file that is associated with the video at the URL passed.

        :param str url: URL of the video to be downloaded
        :return str: Filename of the file in local storage
        """

        try:
            yt = YouTube(url)
        except RegexMatchError:
            log.error(f"Cannot download file at {url}")
        else:
            stream = yt.streams.first()
            log.info(f"Download for {stream.default_filename} has started")
            start_time = time()
            stream.download()
            end_time = time()
            log.info(f"Download for {stream.default_filename} has finished in {end_time - start_time} seconds")
            return stream.default_filename


class Soundcloud(MusicService):
    """Soundcloud service class."""

    def __init__(self, api_key=None):
        self.name = 'soundcloud'
        if api_key is None:
            self.client_id = '81f430860ad96d8170e3bf1639d4e072'
        else:
            self.client_id = api_key
        self.client = soundcloud.Client(client_id=self.client_id)

    def download(self, url):
        """
        Downloads a MP3 file that is associated with the track at the URL passed.
        
        :param str url: URL of the track to be downloaded
        """

        try:
            track = self.client.get('/resolve', url=url)
        except HTTPError:
            log.error(f"{url} is not a Soundcloud URL.")
            return
        r = requests.get(self.client.get(track.stream_url, allow_redirects=False).location, stream=True)
        total_size = int(r.headers['content-length'])
        chunk_size = 1000000
        file_name = track.title + '.mp3'
        with open(file_name, 'wb') as f:
            for data in tqdm(r.iter_content(chunk_size), desc=track.title, total=total_size / chunk_size, unit='MB', file=sys.stdout):
                f.write(data)
        return file_name


class GoogleDrive(StorageService):
    """Google Drive service class."""

    def __init__(self):
        self.name = 'google drive'
        self.connection = None

    def connect(self):
        """Creates connection to the Google Drive API, sets the connection attribute to make requests, and creates the Music folder if it doesn't exist."""

        SCOPES = 'https://www.googleapis.com/auth/drive'
        store = file.Storage('drive_credentials.json')
        creds = store.get()
        if not creds or creds.invalid:
            try:
                flow = client.flow_from_clientsecrets('client_secret.json', SCOPES)
            except InvalidClientSecretsError:
                log.error('ERROR: Could not find client_secret.json in current directory, please obtain it from the API console.')
                return
            creds = tools.run_flow(flow, store)
        self.connection = build('drive', 'v3', http=creds.authorize(Http()))

        response = self.connection.files().list(q="name='Music' and mimeType='application/vnd.google-apps.folder' and trashed=false").execute()
        try:
            folder_id = response.get('files', [])[0]['id']
        except IndexError:
            log.warning('Music folder is missing. Creating it.')
            folder_metadata = {'name': 'Music', 'mimeType': 'application/vnd.google-apps.folder'}
            folder = self.connection.files().create(body=folder_metadata, fields='id').execute()

    def upload(self, file_name):
        """
        Uploads the file associated with the file_name passed to Google Drive in the Music folder.

        :param str file_name: Filename of the file to be uploaded
        :return str: Original filename passed as an argument (in order for the worker to send it to the delete queue)
        """

        response = self.connection.files().list(q="name='Music' and mimeType='application/vnd.google-apps.folder' and trashed=false").execute()
        folder_id = response.get('files', [])[0]['id']
        file_metadata = {'name': file_name, 'parents': [folder_id]}
        media = MediaFileUpload(file_name, mimetype='audio/mpeg')
        
        log.info(f"Upload for {file_name} has started")
        start_time = time()
        self.connection.files().create(body=file_metadata, media_body=media, fields='id').execute()
        end_time = time()
        log.info(f"Upload for {file_name} has finished in {end_time - start_time} seconds")

        return file_name


class LocalStorage(StorageService):
    """Local Storage service class."""

    def __init__(self, custom_path=None):
        self.name = 'local'
        if os.path.exists(custom_path):
            self.music_folder = custom_path
        else:
            log.warning(f"Custom path '{custom_path}' doesn't exist. Using default path.")
            self.music_folder = None

    def connect(self):
        """Initializes the connection attribute with the path to the user home folder's Music folder, and creates it if it doesn't exist."""

        if self.music_folder is None:
            music_folder = os.path.join(os.path.expanduser('~'), 'Music')
            if not os.path.exists(music_folder):
                os.makedirs(music_folder)
            self.music_folder = music_folder

    def upload(self, file_name):
        """
        Moves the file associated with the file_name passed to the Music folder in the local storage.
        
        :param str file_name: Filename of the file to be uploaded
        """
        
        log.info(f"Upload for {file_name} has started")
        start_time = time()
        os.rename(file_name, os.path.join(self.music_folder, file_name))
        end_time = time()
        log.info(f"Upload for {file_name} has finished in {end_time - start_time} seconds")



# -*- coding: utf-8 -*-

from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='music2storage',
      version='0.2.1',
      description='Downloads music from a service and stores it in a cloud or local storage.',
      long_description='Currently the only music services are Youtube and Soundcloud, and the only storages are Google Drive and Local Storage. Dropbox is planned to be added soon.',
      url='https://github.com/Music-Moo/music2storage',
      author='Radu Raicea',
      author_email='radu@raicea.com',
      license='GPLv3',
      packages=['music2storage'],
      install_requires=[
          'ffmpy',
          'google-api-python-client',
          'pytube',
          'requests',
          'tqdm',
          'soundcloud'
      ],
      include_package_data=True,
      python_requires=">=3.6",
      zip_safe=False,
      test_suite='nose.collector',
      tests_require=['nose'],
      classifiers=[
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation :: CPython'
      ])



# -*- coding: utf-8 -*-

import os
import subprocess
from time import time

from ffmpy import FFmpeg, FFRuntimeError

from music2storage import log


def convert_to_mp3(file_name, delete_queue):
    """
    Converts the file associated with the file_name passed into a MP3 file.

    :param str file_name: Filename of the original file in local storage
    :param Queue delete_queue: Delete queue to add the original file to after conversion is done
    :return str: Filename of the new file in local storage
    """


    file = os.path.splitext(file_name)

    if file[1] == '.mp3':
        log.info(f"{file_name} is already a MP3 file, no conversion needed.")
        return file_name

    new_file_name = file[0] + '.mp3'

    ff = FFmpeg(
        inputs={file_name: None},
        outputs={new_file_name: None}
    )

    log.info(f"Conversion for {file_name} has started")
    start_time = time()
    try:
        ff.run(stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except FFRuntimeError:
        os.remove(new_file_name)
        ff.run(stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    end_time = time()
    log.info(f"Conversion for {file_name} has finished in {end_time - start_time} seconds")

    delete_queue.put(file_name)
    return new_file_name


def delete_local_file(file_name):
    """
    Deletes the file associated with the file_name passed from local storage.
    
    :param str file_name: Filename of the file to be deleted
    :return str: Filename of the file that was just deleted
    """

    try:
        os.remove(file_name)
        log.info(f"Deletion for {file_name} has finished")
        return file_name
    except OSError:
        pass
